/* App Switcher */

.switcher-popup {
  padding: 8px;
  spacing: $base_spacing * 4;
}

// switcher onscreen panel
.switcher-list {
  @extend %osd_panel;

  .item-box {
    padding: 8px;
    border-radius: $base_border_radius + 1px;
    border: 1px solid transparent;

    &:outlined {
      border: 1px solid darken($borders_color,5%);
      background-color: transparentize($osd_fg_color, 0.9);
      box-shadow: inset 0 2px 2px 0 rgba(0,0,0,0.4);
    }

    &:selected {
      background-color: transparentize($osd_fg_color, 0.9);
      color: $osd_fg_color;
    }
  }

  // window thumbnails
  .thumbnail-box {
    padding: 2px;
    spacing: $base_spacing;
  }

  .thumbnail {
    width: 256px;
  }

  .separator {
    width: 1px;
    background: $borders_color;
  }

  .switcher-list-item-container {
    spacing: $base_spacing * 2;
  }
}

.switcher-arrow {
  border-color: rgba(0,0,0,0);
  color: transparentize($fg_color,0.2);
  &:highlighted {
    color: $fg_color;
  }
}

// Input Source Switcher
.input-source-switcher-symbol {
  font-size: 34pt;
  width: 96px;
  height: 96px;
}

// Window cycler highlight
.cycler-highlight {
  border: 5px solid $selected_bg_color;
}
