/* Workspace Switcher */
.workspace-switcher-group {
  padding: $base_padding * 2;
}

.workspace-switcher-container {
  @extend %osd_panel;
}

.workspace-switcher {
  background: transparent;
  border: none;
  border-radius: 0;
  padding: 0;
  spacing: $base_spacing * 2;
}

.ws-switcher-box {
  // background: transparent;
  background: transparent;
  height: 50px;
  background-size: 32px;
  border: 1px solid transparentize($osd_fg_color,0.9);
  border-radius: $base_border_radius + 3px;
}

// active workspace in the switcher
.ws-switcher-active-up,
.ws-switcher-active-down,
.ws-switcher-active-left,
.ws-switcher-active-right {
  background-color: $selected_bg_color;
  border: 1px solid if($variant=='light', darken($selected_bg_color, 8%), lighten($selected_bg_color, 5%));
  border-radius: $base_border_radius + 3px;
  color: $selected_fg_color;
}
